import asyncio
import json
import logging
import types
import unittest

import aiohttp
from kutana import (Attachment, VKEnvironment, VKManager, VKRequest,
                    VKResponse, set_logger_level)


class TestManagerVk(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        logging.disable(logging.INFO)

    def tearDown(self):
        logging.disable(logging.DEBUG)

    def test_vk_exceptions(self):
        with self.assertRaises(ValueError):
            VKManager("")

    def test_vk_no_peer_id(self):
        env = VKEnvironment(VKManager("token"), None)

        res = self.loop.run_until_complete(
            env.reply("message")
        )

        self.assertEqual(res, ())

    def test_vk_manager_raw_request(self):
        mngr = VKManager("token")

        class FakeSession:
            def post(self, url, data):
                class FakePost:
                    def __init__(self, url, data):
                        pass

                    async def __aenter__(self):
                        class FakeResponse:
                            status = 200

                            async def text(self):
                                return json.dumps(
                                    {"error":{"error_code":5,"error_msg":"User authori"
                                    "zation failed: invalid access_token (4).",
                                    "request_params":[{"key":"oauth","value":"1"},
                                    {"key":"method","value":"any.method"},{"key":"v",
                                    "value":"5.80"},{"key":"a1","value":"v1"},{
                                    "key":"a2","value":"v2"}]}}
                                )

                        return FakeResponse()

                    async def __aexit__(self, exc_type, exc, tb):
                        pass

                return FakePost(url, data)

            async def close(self):
                pass

        async def test():
            mngr.session = FakeSession()

            set_logger_level(logging.CRITICAL)

            response = await mngr.raw_request("any.method", a1="v1", a2="v2")

            set_logger_level(logging.ERROR)

            self.assertEqual(response.errors[0][1]["error_code"], 5)

        self.loop.run_until_complete(
            test()
        )

        self.loop.run_until_complete(
            mngr.dispose()
        )

    def test_vk_manager_request(self):
        mngr = VKManager("token")

        response = self.loop.run_until_complete(
            mngr.request("fake.method", _timeout=0)
        )

        self.assertTrue(response.error)
        self.assertEqual(len(mngr.requests_queue), 1)
        self.assertEqual(mngr.requests_queue[0].method, "fake.method")
        self.assertEqual(mngr.requests_queue[0].kwargs, {})

    def test_vk_manager_send_message(self):
        mngr = VKManager("token")

        responses = self.loop.run_until_complete(
            mngr.send_message(
                "text for message", 0, ["attachment"], random_id=0, _timeout=0
            )
        )

        response = responses[0]

        self.assertTrue(response.error)
        self.assertEqual(len(mngr.requests_queue), 1)
        self.assertEqual(mngr.requests_queue[0].method, "messages.send")
        self.assertEqual(mngr.requests_queue[0].kwargs, {
            "message": "text for message",
            "attachment": "attachment,",
            "peer_id": 0, "random_id": 0
        })

    def test_vk_manager_send_message_long(self):
        mngr = VKManager("token")

        responses = self.loop.run_until_complete(
            mngr.send_message(
                "a" * 6000, 0, random_id=0, _timeout=0
            )
        )

        self.assertEqual(len(responses), 2)

        self.assertTrue(responses[0].error)
        self.assertTrue(responses[1].error)

        self.assertEqual(len(mngr.requests_queue), 2)

        self.assertEqual(mngr.requests_queue[0].method, "messages.send")
        self.assertEqual(mngr.requests_queue[0].kwargs["message"], "a" * 4096)
        self.assertTrue(mngr.requests_queue[0].kwargs["random_id"])

        self.assertEqual(mngr.requests_queue[1].method, "messages.send")
        self.assertEqual(mngr.requests_queue[1].kwargs["message"],
                         "a" * (6000 - 4096))
        self.assertEqual(mngr.requests_queue[1].kwargs["random_id"], 0)

    def test_vk_manager_send_message_attachment(self):
        mngr = VKManager("token")

        attachment = Attachment("photo", 1, 0, None, None, None)

        responses = self.loop.run_until_complete(
            mngr.send_message(
                "text for message", 0, attachment, _timeout=0
            )
        )

        response = responses[0]

        self.assertTrue(response.error)
        self.assertEqual(len(mngr.requests_queue), 1)
        self.assertEqual(mngr.requests_queue[0].method, "messages.send")
        self.assertIsNotNone(mngr.requests_queue[0].kwargs.get("random_id"))
        self.assertEqual(mngr.requests_queue[0].kwargs, {
            "message": "text for message",
            "attachment": "photo0_1,",
            "peer_id": 0,
            "random_id": mngr.requests_queue[0].kwargs.get("random_id")
        })

    def test_vk_manager_create_attachment(self):
        test_attachment_raw = {
            "id": 13, "album_id": 13, "owner_id": 1, "sizes":
            [{"type": "s", "url": "url", "width": 0, "height": 0}],
            "text": "", "date": 0, "post_id": 0, "likes":
            {"user_likes": 0, "count": 577904}, "comments": {"count": 421},
            "can_comment": 1, "can_repost": 1, "tags": {"count": 0}
        }

        attachment = VKManager.create_attachment(
            test_attachment_raw, "photo"
        )

        self.assertEqual(attachment.id, 13)
        self.assertEqual(attachment.type, "photo")
        self.assertEqual(attachment.owner_id, 1)
        self.assertEqual(attachment.access_key, None)
        self.assertEqual(attachment.link, "url")
        self.assertEqual(attachment.raw_attachment, test_attachment_raw)

    def test_vk_manager_create_attachment_no_type(self):
        test_attachment_raw = {
            "type": "photo",
            "photo": {
                "id": 13, "album_id": 13, "owner_id": 1, "sizes":
                [{"type": "s", "url": "url2", "width": 0, "height": 0},
                {"type": "m", "url": "url", "width": 0, "height": 0}],
                "text": "", "date": 0, "post_id": 0, "likes":
                {"user_likes": 0, "count": 577904}, "comments": {"count": 421},
                "can_comment": 1, "can_repost": 1, "tags": {"count": 0}
            }
        }

        attachment = VKManager.create_attachment(
            test_attachment_raw
        )

        self.assertEqual(attachment.id, 13)
        self.assertEqual(attachment.type, "photo")
        self.assertEqual(attachment.owner_id, 1)
        self.assertEqual(attachment.access_key, None)
        self.assertEqual(attachment.link, "url")
        self.assertEqual(attachment.raw_attachment, test_attachment_raw)

    def test_vk_manager_create_attachment_doc(self):
        test_attachment_raw = {
            "id": 13, "owner_id": 1, "title": "rrrrr.png", "size": 119900,
            "ext": "png", "url": "url", "date": 0, "type": 4,
        }

        attachment = VKManager.create_attachment(
            test_attachment_raw, "doc"
        )

        self.assertEqual(attachment.id, 13)
        self.assertEqual(attachment.type, "doc")
        self.assertEqual(attachment.owner_id, 1)
        self.assertEqual(attachment.access_key, None)
        self.assertEqual(attachment.link, "url")
        self.assertEqual(attachment.raw_attachment, test_attachment_raw)

    def test_vk_manager_create_attachment_doc_no_link(self):
        test_attachment_raw = {
            "id": 13, "owner_id": 1, "title": "rrrrr.png", "size": 119900,
            "ext": "png", "date": 0, "type": 4,
        }

        attachment = VKManager.create_attachment(
            test_attachment_raw, "doc"
        )

        self.assertEqual(attachment.id, 13)
        self.assertEqual(attachment.type, "doc")
        self.assertEqual(attachment.owner_id, 1)
        self.assertEqual(attachment.access_key, None)
        self.assertEqual(attachment.link, None)
        self.assertEqual(attachment.raw_attachment, test_attachment_raw)

    def test_vk_startup(self):
        mngr = VKManager("token")

        bg_coroutines = []

        asyncio_ensure_future = asyncio.ensure_future

        def ensure(task):
            _task = asyncio_ensure_future(task)

            bg_coroutines.append(_task)

            return _task

        asyncio.ensure_future = ensure

        self.loop.run_until_complete(
            mngr.startup(None)
        )

        asyncio.ensure_future = asyncio_ensure_future

        mngr.running = False

        self.loop.run_until_complete(bg_coroutines[0])

        self.assertEqual(len(bg_coroutines), 1)

    def test_vk_environment(self):
        mngr = VKManager("token")
        env = self.loop.run_until_complete(
            mngr.get_environment({"object": {"peer_id": 1}})
        )

        self.assertEqual(env.peer_id, 1)
        self.assertEqual(env.manager, mngr)

    def test_vk_update_longpoll_data(self):
        mngr = VKManager("token")

        async def raw_request(self, method, **kwargs):
            return VKResponse(
                False, (), {"key": "key", "server": "server", "ts": "0"}
            )

        mngr.raw_request = types.MethodType(raw_request, mngr)

        self.loop.run_until_complete(
            mngr.update_longpoll_data()
        )

        self.loop.run_until_complete(
            mngr.dispose()
        )

        self.assertEqual(
            mngr.longpoll,
            {"key": "key", "server": "server", "ts": "0"}
        )

    def test_vk_update_longpoll_data_exception(self):
        mngr = VKManager("token")

        async def raw_request(self, method, **kwargs):
            return VKResponse(True, (), "")

        mngr.raw_request = types.MethodType(raw_request, mngr)

        with self.assertRaises(ValueError):
            self.loop.run_until_complete(
                mngr.update_longpoll_data()
            )

        self.loop.run_until_complete(
            mngr.dispose()
        )

        self.assertEqual(mngr.longpoll, None)

    def test_vk_receiver(self):
        mngr = VKManager("token")

        async def prepare():
            mngr.longpoll = {
                "ts": 0, "server": "server", "key": "key"
            }

            class FakeSession:
                def post(self, url):
                    class FakePost:
                        def __init__(self, url):
                            pass

                        async def __aenter__(self):
                            class FakeResponse:
                                async def json(self):
                                    return {
                                        "ts": "4",
                                        "updates": [
                                            {"type": "type", "object": "object"},
                                            "update2"
                                        ]
                                    }

                            return FakeResponse()

                        async def __aexit__(self, exc_type, exc, tb):
                            pass

                    return FakePost(url)

                async def close(self):
                    pass

            mngr.session = FakeSession()

        self.loop.run_until_complete(prepare())

        updates = self.loop.run_until_complete(mngr.receiver())

        self.assertEqual(updates, [{"type": "type", "object": "object"}])
        self.assertEqual(mngr.longpoll["ts"], "4")

        self.loop.run_until_complete(mngr.dispose())


    def test_vk_receiver_failed(self):
        mngr = VKManager("token")

        async def prepare():
            mngr.longpoll = {
                "ts": 0, "server": "server", "key": "key"
            }

            class FakeSession:
                def post(self, url):
                    class FakePost:
                        def __init__(self, url):
                            pass

                        async def __aenter__(self):
                            class FakeResponse:
                                async def json(self):
                                    return {"failed": 2}

                            return FakeResponse()

                        async def __aexit__(self, exc_type, exc, tb):
                            pass

                    return FakePost(url)

                async def close(self):
                    pass

            mngr.session = FakeSession()

            async def update_longpoll_data(self):
                self.longpoll = "updated"

            mngr.update_longpoll_data = types.MethodType(
                update_longpoll_data, mngr
            )

        self.loop.run_until_complete(prepare())

        updates = self.loop.run_until_complete(mngr.receiver())

        self.assertEqual(updates, ())
        self.assertEqual(mngr.longpoll, "updated")

        self.loop.run_until_complete(mngr.dispose())

    def test_vk_get_receiver_coroutine_function(self):
        mngr = VKManager("token")

        async def raw_request(self, method, **kwargs):
            return VKResponse(False, (), [{"id": 1, "name": "keks"}])

        mngr.raw_request = types.MethodType(raw_request, mngr)

        async def update_longpoll_data(self):
            self.longpoll = "updated"

        mngr.update_longpoll_data = types.MethodType(
            update_longpoll_data, mngr
        )

        receiver = self.loop.run_until_complete(
            mngr.get_receiver_coroutine_function()
        )

        self.assertEqual(mngr.longpoll, "updated")
        self.assertEqual(mngr.receiver, receiver)
        self.assertEqual(mngr.group_id, 1)

        self.loop.run_until_complete(mngr.dispose())

    def test_vk_exec_loop(self):
        mngr = VKManager("token")

        mngr.running = True

        async def raw_request(_, method, **kwargs):
            self.assertEqual(method, "execute")
            self.assertEqual(kwargs["code"], "return [API.fake.method({}),];")

            mngr.running = False

            return VKResponse(False, (), {"response": ["response"]})

        mngr.raw_request = types.MethodType(raw_request, mngr)

        req = VKRequest("fake.method", {})

        mngr.requests_queue.append(req)

        tasks = []
        def ensure(task):
            _task = asyncio.ensure_future(task, loop=self.loop)
            tasks.append(_task)
            return _task

        self.loop.run_until_complete(
            mngr._exec_loop()
        )

        self.loop.run_until_complete(asyncio.gather(*tasks))

        response = self.loop.run_until_complete(req)

        self.assertFalse(response.error)
        self.assertEqual(response.response, "response")

    def test_vk_exec_loop_error(self):
        mngr = VKManager("token")

        mngr.running = True

        async def raw_request(_, method, **kwargs):
            self.assertEqual(method, "execute")
            self.assertEqual(
                kwargs["code"],
                'return [API.users.get({"user_ids": "sdfsdfsdfsdfsdf"}),];'
            )

            mngr.running = False

            raw_response = {
                "response": [False],
                "execute_errors": [
                    {
                        "method": "users.get",
                        "error_code": 113,
                        "error_msg": "Invalid user id"
                    }, {
                        "method": "execute",
                        "error_code": 113,
                        "error_msg": "Invalid user id"
                    }
                ]
            }

            return VKResponse(
                True,
                (
                    ("VK_req", raw_response.get("error", "")),
                    ("VK_exe", raw_response.get("execute_errors", ""))
                ),
                raw_response.get("response", "")
            )

        mngr.raw_request = types.MethodType(raw_request, mngr)

        req = VKRequest("users.get", {"user_ids": "sdfsdfsdfsdfsdf"})

        mngr.requests_queue.append(req)

        logging.disable(logging.ERROR)

        tasks = []
        def ensure(task):
            _task = asyncio.ensure_future(task, loop=self.loop)
            tasks.append(_task)
            return _task

        self.loop.run_until_complete(
            mngr._exec_loop()
        )

        self.loop.run_until_complete(asyncio.gather(*tasks))

        logging.disable(logging.INFO)

        response = self.loop.run_until_complete(req)

        self.assertTrue(response.error)
        self.assertEqual(response.errors[0][0], "VK_req")
        self.assertEqual(response.response, "")

    @staticmethod
    def get_fake_manager(actions):
        class FakeManager:
            group_id = 10

            @staticmethod
            def create_attachment(att, typ):
                actions.append((att, typ))

                return "file"

            async def request(self, method, **kwargs):
                if method == "docs.getMessagesUploadServer":
                    actions.append((method, kwargs))
                    return VKResponse(False, (), {"upload_url": "url1"})

                if method == "docs.getWallUploadServer":
                    actions.append((method, kwargs))
                    return VKResponse(False, (), {"upload_url": "url2"})

                if method == "photos.getMessagesUploadServer":
                    actions.append((method, kwargs))
                    return VKResponse(False, (), {"upload_url": "url3"})

                if method == "photos.saveMessagesPhoto":
                    actions.append((method, kwargs))
                    return VKResponse(False, (), ["attachment"])

                if method == "docs.save":
                    actions.append((method, kwargs))
                    return VKResponse(False, (), "attachment")

                raise ValueError

        return FakeManager()

    def test_vk_upload_doc(self):
        actions = []

        env = VKEnvironment(self.get_fake_manager(actions), None)

        async def _upload_file_to_vk(self, url, data):
            return {"doc": "upload"}

        env.replace_method("_upload_file_to_vk", _upload_file_to_vk)

        attachment = self.loop.run_until_complete(
            env.upload_doc(b"file")
        )

        self.assertEqual(attachment, "file")

        self.assertEqual(len(actions), 3)

        self.assertEqual(
            actions[0], ('docs.getWallUploadServer', {'group_id': 10})
        )

        self.assertEqual(
            actions[1], ('docs.save', {'doc': 'upload'})
        )

        self.assertEqual(
            actions[2], ('attachment', 'doc')
        )

    def test_vk_upload_doc_w_peer_id(self):
        actions = []

        env = VKEnvironment(self.get_fake_manager(actions), None)

        async def _upload_file_to_vk(self, url, data):
            return {"doc": "upload"}

        env.replace_method("_upload_file_to_vk", _upload_file_to_vk)

        attachment = self.loop.run_until_complete(
            env.upload_doc(b"file", peer_id=2)
        )

        self.assertEqual(attachment, "file")

        self.assertEqual(len(actions), 3)

        self.assertEqual(
            actions[0],
            ('docs.getMessagesUploadServer', {'peer_id': 2, 'type': 'doc'})
        )

        self.assertEqual(
            actions[1], ('docs.save', {'doc': 'upload'})
        )

        self.assertEqual(
            actions[2], ('attachment', 'doc')
        )

    def test_vk_upload_photo(self):
        actions = []

        env = VKEnvironment(self.get_fake_manager(actions), None)

        async def _upload_file_to_vk(self, url, data):
            return {"doc": "upload"}

        env.replace_method("_upload_file_to_vk", _upload_file_to_vk)

        attachment = self.loop.run_until_complete(
            env.upload_photo(b"file")
        )

        self.assertEqual(attachment, "file")

        self.assertEqual(len(actions), 3)

        self.assertEqual(
            actions[0], ('photos.getMessagesUploadServer', {'peer_id': None})
        )

        self.assertEqual(
            actions[1], ('photos.saveMessagesPhoto', {'doc': 'upload'})
        )

        self.assertEqual(
            actions[2], ('attachment', 'photo')
        )

    def test_vk_upload_photo_w_peer_id(self):
        actions = []

        env = VKEnvironment(self.get_fake_manager(actions), None)

        async def _upload_file_to_vk(self, url, data):
            return {"doc": "upload"}

        env.replace_method("_upload_file_to_vk", _upload_file_to_vk)

        attachment = self.loop.run_until_complete(
            env.upload_photo(b"file", peer_id=2)
        )

        self.assertEqual(attachment, "file")

        self.assertEqual(len(actions), 3)

        self.assertEqual(
            actions[0],
            ('photos.getMessagesUploadServer', {'peer_id': 2})
        )

        self.assertEqual(
            actions[1], ('photos.saveMessagesPhoto', {'doc': 'upload'})
        )

        self.assertEqual(
            actions[2], ('attachment', 'photo')
        )

    def test_upload_file_to_vk(self):
        class FakePost:
            def __init__(self, url, data):
                pass

            async def __aenter__(self):
                class FakeResponse:
                    status = 200

                    async def text(self):
                        return json.dumps({"test": "test"})

                return FakeResponse()

            async def __aexit__(self, exc_type, exc, tb):
                pass

        class Session:
            post = FakePost

        class Manager:
            session = Session()

        env = VKEnvironment(Manager(), None)

        result = self.loop.run_until_complete(
            env._upload_file_to_vk("nice_url", {"nice": "data"})
        )

        self.assertEqual(result, {"test": "test"})

    def test_upload_file_to_vk_error(self):
        class FakePost:
            def __init__(self, url, data):
                pass

            async def __aenter__(self):
                class FakeResponse:
                    status = 200

                    async def text(self):
                        return json.dumps({"error": "exists"})

                return FakeResponse()

            async def __aexit__(self, exc_type, exc, tb):
                pass

        class Session:
            post = FakePost

        class Manager:
            session = Session()

        env = VKEnvironment(Manager())

        result = self.loop.run_until_complete(
            env._upload_file_to_vk("nice_url", {"nice": "data"})
        )

        self.assertEqual(result, None)

    def test_vk_get_file_from_attachment_empty(self):
        mngr = VKManager("token")

        async def test():
            mngr.session = aiohttp.ClientSession()

            env = VKEnvironment(mngr)

            res = await env.get_file_from_attachment(None)

            self.assertEqual(res, None)

            res = await env.get_file_from_attachment(
                Attachment("photo", 13, 1, None, None, {})
            )

            self.assertEqual(res, None)

        self.loop.run_until_complete(test())

        self.loop.run_until_complete(mngr.dispose())



"""Manager and environment for debug purposes."""

import time

from kutana.environment import Environment
from kutana.exceptions import ExitException
from kutana.manager.manager import Manager
from kutana.plugin import Message


class DebugEnvironment(Environment):

    """Environment for :class:`.DebugManager`."""

    async def upload_doc(self, file, **kwargs):
        """Return first argument."""

        return file

    async def upload_photo(self, file, **kwargs):
        """Return first argument."""

        return file

    async def get_file_from_attachment(self, attachment):
        """Return first argument."""

        return attachment


class DebugManager(Manager):

    """Shoots target texts once and contains replied data."""

    _type = "debug"

    def __init__(self, *texts):
        self.replies = []
        self.replies_all = {}

        self.queue = list(texts)
        self.dead = False

    async def request(self, method, **kwargs):  # pylint: disable=W0613
        """Placeholder for debug's "request" method."""

        return None

    @staticmethod
    async def create_message(update):
        # update = (sender_id, message)
        # or
        # update = (message_from_user_with_id_1)

        if not update:
            return None

        if isinstance(update, str):
            return Message(
                str(update), (), 1, 1, time.time(), (1, str(update))
            )

        return Message(
            update[1], (), update[0], update[0], time.time(), update
        )

    async def send_message(self, message, peer_id, attachment=None, **kwargs):
        """
        Send message to specified user or user with id 1.

        :param message: message to send
        :param peer_id: recipient's id
        :param attachment: optional attachment or list of attachments to
            reply with
        :param kwargs: optional parameters (means nothing)
        :returns: None
        """

        peer_id = 1 if peer_id is None else peer_id

        if peer_id not in self.replies_all:
            self.replies_all[peer_id] = []

        replies_all = self.replies_all[peer_id]

        if message:
            replies_all.append(message)

            if peer_id == 1:
                self.replies.append(message)

        if attachment:
            if not isinstance(attachment, (list, tuple)):
                attachment = [attachment]

            replies_all.extend(attachment)

            if peer_id == 1:
                self.replies.extend(attachment)

    async def get_environment(self, update):
        peer_id = update[0] if isinstance(update, (list, tuple)) else 1

        return DebugEnvironment(self, peer_id=peer_id)

    async def get_receiver_coroutine_function(self):
        async def rec():
            if self.dead:
                raise ExitException

            self.dead = True

            return self.queue

        return rec

    async def startup(self, application):
        pass

    async def dispose(self):
        pass



"""Functions for loading and extracting plugins from python files."""

import importlib.util
import os
import re

from .logger import logger
from .plugin import Plugin
from .utils import is_list_or_tuple


def import_module(name, path):
    """
    Import module from specified path with specified name.

    :param name: module's name
    :param path: path to module's file
    :returns: imported module
    """

    spec = importlib.util.spec_from_file_location(name, os.path.abspath(path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def load_plugins_from_file(path, verbose=False):
    """
    Load plugins from specified path. Plugins can be in module-level
    variable "plugin" and in module-level variable "plugins" with list of
    plugins.

    :param path: path to plugin's file
    :returns: loaded module or None if no plugin found
    """

    module = import_module(path, path)

    plugins = []

    possible_plugins = []

    if hasattr(module, "plugin") and isinstance(module.plugin, Plugin):
        possible_plugins.append(module.plugin)

    if hasattr(module, "plugins") and is_list_or_tuple(module.plugins):
        possible_plugins.extend(module.plugins)

    for plugin in possible_plugins:
        if plugin and isinstance(plugin, Plugin):
            plugins.append(plugin)

    if plugins:
        count = len(plugins)

        if count == 1:
            logger.info('Loaded plugin from "%s"', path)

        else:
            logger.info('Loaded %d plugins from "%s"', count, path)

    elif verbose:  # pragma: no cover
        logger.warning("No plugins found in \"%s\"", path)

    return plugins


def load_plugins(folder):
    """
    Import all plugins from target folder recursively.

    :param folder: path to target folder
    :returns: list of loaded plugins
    """

    plugins_list = []

    for name in os.listdir(folder):
        path = os.path.join(folder, name)

        if os.path.isdir(path):
            plugins_list += load_plugins(path)

        elif re.match(r"^[^_].*\.py$", name):
            plugins_list += load_plugins_from_file(path)

    return plugins_list



import unittest
from contextlib import contextmanager

from kutana import DebugManager, Kutana, Plugin


class KutanaTest(unittest.TestCase):
    def setUp(self):
        self.app = None
        self.target = []

    @contextmanager
    def debug_manager(self, queue):
        if self.app is None:
            self.app = Kutana()

        self.manager = DebugManager(*queue)

        self.app.add_manager(self.manager)

        self.plugin = Plugin()

        self.plugins = [self.plugin]

        try:
            yield self.plugin
        finally:
            for plugin in self.plugins:
                self.app.register_plugins([plugin])

            self.app.run()

        self.assertEqual(self.target, self.manager.replies)



"""Base for managers."""


class Manager():

    """Base for manager."""

    _type = ""

    def get_type(self):
        """Return this manager's type."""

        return self._type

    @staticmethod
    def split_large_text(text):
        """
        Split text into chunks with length of 4096.

        :param text: text for splitting
        :returns: tuple of chunks
        """

        if len(text) < 4096:
            return (text,)

        return tuple(
            text[i : i + 4096] for i in range(0, len(text), 4096)
        )

    async def get_environment(self, update):
        """
        Create and return manager's environment for update processing.

        :param update: manager's service raw update
        :returns: :class:`.Environment` for manager's service
        """

        raise NotImplementedError

    async def create_message(self, update):
        """
        Create and return :class:`.Message` from raw_update.

        :param update: manager's service raw update
        :returns: :class:`.Message` or None if message can't be created
        """

        raise NotImplementedError

    async def request(self, method, **kwargs):
        """
        Perform request to manager's service and return result.

        :returns: response
        """

        raise NotImplementedError

    async def send_message(self, message, peer_id, attachment=None, **kwargs):
        """
        Send message to target user with "peer_id" with parameters. This can
        vary in managers' implementations.

        :param message: text to send
        :param peer_id: target recipient
        :param attachment: list of instances of :class:`.Attachment`
        :parma kwargs: arguments to send to manager's methods
        :returns: list of responses
        """

        raise NotImplementedError

    async def get_receiver_coroutine_function(self):
        """
        Collect coroutine for receiving updated from service.

        :returns: coroutine for receiving updated from service
        """

        raise NotImplementedError

    async def startup(self, application):
        """
        Start coroutines and prepare everything else needed for
        manager to work.

        :param app: application
        """

        raise NotImplementedError

    async def dispose(self):
        """Free used resources."""

        raise NotImplementedError



"""Environment for :class:`.VKManager`."""

import json
import aiohttp
from kutana.environment import Environment


class VKEnvironment(Environment):

    """Environment for :class:`.VKManager`"""

    async def _upload_file_to_vk(self, upload_url, data):
        upload_result_text = None

        async with self.manager.session.post(upload_url, data=data) as resp:
            if resp.status == 200:
                upload_result_text = await resp.text()

        try:
            upload_result = json.loads(upload_result_text)

            if "error" in upload_result:
                raise RuntimeError

        except RuntimeError:
            upload_result = None

        return upload_result

    async def reply(self, message, attachment=None, **kwargs):
        """
        Reply to currently processed message.

        :param message: message to reply with
        :param attachment: optional attachment or list of attachments to
            reply with
        :param kwargs: arguments for vkontakte's "messages.send"
        :returns: list with results of sending messages
        """

        return await self.manager.send_message(
            message, self.peer_id, attachment, **kwargs
        )

    async def upload_doc(self, file, **kwargs):
        """
        Upload file to be sent with :func:`.send_message`
        (or :func:`.reply`) as document. If you passed "peer_id", vkontakte's
        method "docs.getWallUploadServer" will be used.

        :param file: document as file or bytes
        :param kwargs: arguments for vkontakte's methods and optional
            "filename"
        :returns: :class:`.Attachment` or None
        """

        if kwargs.get("peer_id") is None:
            peer_id = self.peer_id

        else:
            peer_id = kwargs["peer_id"]

        if peer_id:
            upload_data = await self.manager.request(
                "docs.getMessagesUploadServer",
                peer_id=peer_id,
                type=kwargs.get("doctype", "doc")
            )

        else:
            upload_data = await self.manager.request(
                "docs.getWallUploadServer",
                group_id=kwargs.get("group_id") or self.manager.group_id
            )

        if upload_data.error:
            return None

        upload_url = upload_data.response["upload_url"]

        data = aiohttp.FormData()
        data.add_field("file", file, filename=kwargs.get("filename"))

        upload_result = await self._upload_file_to_vk(upload_url, data)

        if not upload_result:
            return None

        attachments = await self.manager.request(
            "docs.save", **upload_result
        )

        if attachments.error:
            return None

        return self.manager.create_attachment(
            attachments.response, "doc"
        )

    async def upload_photo(self, file, **kwargs):
        """
        Upload file to be sent with :func:`.send_message`
        (or :func:`.reply`) as photo. If "peer_id" was passed, file will be
        uploaded for user with "peer_id".

        :param file: photo as file or bytes
        :param kwargs: arguments for vkontakte's methods
        :returns: :class:`.Attachment` or None
        """

        if kwargs.get("peer_id") is None:
            peer_id = self.peer_id

        else:
            peer_id = kwargs.get("peer_id")

        upload_data = await self.manager.request(
            "photos.getMessagesUploadServer", peer_id=peer_id
        )

        if upload_data.error:
            return None

        upload_url = upload_data.response["upload_url"]

        data = aiohttp.FormData()
        data.add_field("photo", file, filename="image.png")

        upload_result = await self._upload_file_to_vk(upload_url, data)

        if not upload_result:
            return None

        attachments = await self.manager.request(
            "photos.saveMessagesPhoto", **upload_result
        )

        if attachments.error:
            return None

        return self.manager.create_attachment(
            attachments.response[0], "photo"
        )

    async def get_file_from_attachment(self, attachment):
        if not attachment or not attachment.link:
            return None

        async with self.manager.session.get(attachment.link) as response:
            return await response.read()



import logging

from kutana import set_logger_level

from testing_tools import KutanaTest


logging.disable(logging.INFO)


class TestKutana(KutanaTest):
    def test_debug(self):
        self.target = ["message"]

        with self.debug_manager(self.target):

            async def new_update(update, env):
                await env.reply(update)

            self.app.register(new_update)

    def test_debug_upload(self):
        self.target = ["file"]

        with self.debug_manager(self.target):

            async def new_update(update, env):
                attachment = await env.upload_doc("file")

                await env.reply("", attachment=attachment)

            self.app.register(new_update)

    def test_priority(self):
        self.called_1 = 0
        self.called_2 = 0

        with self.debug_manager(["message"]):

            async def prc1(update, env):
                self.called_1 += 1
                return "DONE"

            async def prc2(update, env):
                self.called_2 += 1
                return "DONE"

            self.app.register(prc1)
            self.app.register(prc2, priority=10)

        self.assertEqual(self.called_1, 0)
        self.assertEqual(self.called_2, 1)

    def test_exception(self):
        self.called = 0

        self.target = ["Произошла крайне критическая ошибка. "
                       "Сообщите об этом администратору."]

        with self.debug_manager(["message"]):

            async def new_update(update, env):
                async def just_raise(update, env):
                    self.called += 1
                    raise Exception

                env.register_after_processed(just_raise)

                raise Exception

            self.app.register(new_update)

            set_logger_level(logging.CRITICAL)

        set_logger_level(logging.ERROR)

        self.assertEqual(self.called, 1)

    def test_no_default_exception_handle(self):
        self.target = []

        with self.debug_manager(["message"]):

            async def new_update(update, env):
                raise Exception

            self.app.register(new_update)

            set_logger_level(logging.CRITICAL)

        set_logger_level(logging.ERROR)



"""Manager for interacting with VKontakte"""

import asyncio
import json
import re
from collections import namedtuple
from random import random

import aiohttp
from kutana.logger import logger
from kutana.manager.manager import Manager
from kutana.manager.vk.environment import VKEnvironment
from kutana.plugin import Attachment, Message


NAIVE_CACHE = {}


VKResponse = namedtuple(
    "VKResponse",
    "error errors response"
)

VKResponse.__doc__ = """
Response from vkontakte.

:param error: boolean value indicating if error happened
:param errors: array with happened errors
:param response: result of reqeust if no errors happened
"""


class VKRequest(asyncio.Future):

    """Class for queueing requests to VKontakte"""

    def __init__(self, method, kwargs):
        super().__init__()

        self.method = method
        self.kwargs = kwargs


class VKManager(Manager):

    """
    Class for receiving updates from vkontakte.
    Controller requires group's token. You can specify settings for
    groups.setLongPollSettings with argument "longpoll_settings".

    :param token: bot's token
    :param executes_per_second: how many "execute" requests per second bot
        can send
    :param longpoll_settings: values for group's longpoll settings
    :param api_version: vkontakte's api version to use in requests
    """


    _type = "vkontakte"


    def __init__(self, token, executes_per_second=19, longpoll_settings=None,
                 api_version="5.92"):

        if not token:
            raise ValueError("No `token` specified")

        if not longpoll_settings:
            longpoll_settings = {}

        self.session = None
        self.group_id = None
        self.longpoll = None

        self.running = False
        self.requests_queue = []

        self.version = api_version
        self.token = token
        self.execute_pause = 1 / executes_per_second
        self.longpoll_settings = longpoll_settings

        api_url_base = "https://api.vk.com"
        api_url_template = api_url_base + "/method/{{}}?access_token={}&v={}"

        self.api_url = api_url_template.format(self.token, self.version)

        self.longpoll_url_template = "{}?act=a_check&key={}&wait=25&ts={}"

    async def raw_request(self, method, **kwargs):
        """Perform raw api request to vkontakte"""

        if not self.session:
            self.session = aiohttp.ClientSession()

        data = {k: v for k, v in kwargs.items() if v is not None}

        raw_response = {}

        try:
            async with self.session.post(self.api_url.format(method),
                                         data=data) as response:

                raw_response_text = await response.text()

                raw_response = json.loads(raw_response_text)

                logger.debug("Method: %s; Data: %s; Response: %s", method,
                             data, raw_response)

        except (json.JSONDecodeError, aiohttp.ClientError) as e:
            logger.debug("Method: %s; Data: %s; No response", method, data)

            return VKResponse(
                error=True,
                errors=(("Kutana", str(type(e)) + ": " + str(e)),),
                response=""
            )

        if "error" in raw_response:
            return VKResponse(
                error=True,
                errors=(
                    ("VK_req", raw_response.get("error", "")),
                    ("VK_exe", raw_response.get("execute_errors", ""))
                ),
                response=raw_response.get("response", "")
            )

        return VKResponse(
            error=False,
            errors=(
                ("VK_req", raw_response.get("error", "")),
                ("VK_exe", raw_response.get("execute_errors", ""))
            ),
            response=raw_response["response"]
        )

    async def request(self, method, **kwargs):
        """
        Perform request to vkontakte and return result.

        :param method: method to call
        :param timeout: timeout for gettings response from vkontakte
        :param kwargs: parameters for method
        :returns: :class:`.VKResponse`
        """

        timeout = kwargs.pop("_timeout", 180)

        req = VKRequest(
            method,
            kwargs
        )

        self.requests_queue.append(req)

        try:
            return await asyncio.wait_for(req, timeout=timeout)

        except asyncio.TimeoutError:
            return VKResponse(
                error=True,
                errors=(
                    ("Kutana", "Request took too long and was forgotten.")
                ),
                response=""
            )

    async def send_message(self, message, peer_id, attachment=None, **kwargs):
        """
        Send message to target peer_id with parameters.

        :param message: text to send
        :param peer_id: target recipient
        :param attachment: list of :class:`.Attachment` or attachments
            as string
        :parma kwargs: arguments to send to vkontakte's `messages.send`
        :returns: list of responses from telegram
        """

        if peer_id is None:
            return ()

        message_parts = self.split_large_text(message)

        if isinstance(attachment, Attachment):
            attachment = [attachment]

        if isinstance(attachment, (list, tuple)):
            new_attachment = ""

            for a in attachment:
                if isinstance(a, Attachment):
                    new_attachment += (
                        "{}{}_{}".format(a.type, a.owner_id, a.id) +
                        ("_" + a.access_key if a.access_key else "")
                    )

                else:
                    new_attachment += str(a)

                new_attachment += ","

            attachment = new_attachment

        result = []

        for part in message_parts[:-1]:
            temp_kwargs = {
                "random_id": int(random() * 4294967296) - 2147483648
            }

            if kwargs.get("_timeout") is not None:
                temp_kwargs["_timeout"] = kwargs["_timeout"]

            result.append(
                await self.request(
                    "messages.send",
                    message=part,
                    peer_id=peer_id,
                    **temp_kwargs
                )
            )

        if "random_id" not in kwargs:
            kwargs["random_id"] = int(random() * 4294967296) - 2147483648

        result.append(
            await self.request(
                "messages.send",
                message=message_parts[-1],
                peer_id=peer_id,
                attachment=attachment,
                **kwargs
            )
        )

        return result

    async def resolve_screen_name(self, screen_name):
        """Return answer from vkontakte with resolved passed screen name."""

        if screen_name in NAIVE_CACHE:
            return NAIVE_CACHE[screen_name]

        result = await self.request(
            "utils.resolveScreenName",
            screen_name=screen_name
        )

        NAIVE_CACHE[screen_name] = result

        return result

    async def create_message(self, update):
        if update["type"] != "message_new":
            return None

        obj = update["object"]

        text = obj["text"]

        if "conversation_message_id" in obj:
            cursor = 0
            new_text = ""

            for match in re.finditer(r"\[(.+?)\|.+?\]", text):
                resp = await self.resolve_screen_name(match.group(1))

                new_text += text[cursor : match.start()]

                cursor = match.end()

                if not resp.response \
                        or resp.response["object_id"] == update["group_id"]:
                    continue

                new_text += text[match.start() : match.end()]

            new_text += text[cursor :]

            text = new_text.lstrip()

        return Message(
            text,
            tuple(self.create_attachment(a) for a in obj["attachments"]),
            obj.get("from_id"),
            obj.get("peer_id"),
            obj.get("date"),
            update
        )

    @staticmethod
    def create_attachment(attachment, attachment_type=None):
        """
        Create and return :class:`.Attachment` created from passed data. If
        attachment type can't be determined passed "attachment_type"
        well be used.
        """

        if "type" in attachment and attachment["type"] in attachment:
            body = attachment[attachment["type"]]
            attachment_type = attachment["type"]
        else:
            body = attachment

        if "sizes" in body:
            m_s_ind = -1
            m_s_wid = 0

            for i, size in enumerate(body["sizes"]):
                if size["width"] >= m_s_wid:
                    m_s_wid = size["width"]
                    m_s_ind = i

            link = body["sizes"][m_s_ind]["url"]  # src

        elif "url" in body:
            link = body["url"]

        else:
            link = None

        return Attachment(
            attachment_type,
            body.get("id"),
            body.get("owner_id"),
            body.get("access_key"),
            link,
            attachment
        )

    async def get_environment(self, update):
        """
        Returns environment for update.

        :param update: update from vkontakte
        :returns: :class:`.VKEnvironment`
        """

        return VKEnvironment(self, peer_id=update["object"].get("peer_id"))

    async def _exec_perform(self, code, requests):
        result = await self.raw_request("execute", code=code)

        if result.error:
            logger.error(result.errors)

        errors_amount = 0

        if result.errors and result.errors[-1][0] == "VK_exe":
            execute_errors = result.errors[-1][1]

        else:
            execute_errors = []

        for res, req in zip(result.response, requests):
            if res is False:
                known_error = "unknown error"

                if len(execute_errors) > errors_amount:
                    known_error = execute_errors[errors_amount]
                    errors_amount += 1

                response = VKResponse(
                    error=True,
                    errors=(("VK_req", known_error),),
                    response=""
                )

            else:
                response = VKResponse(
                    error=False,
                    errors=(),
                    response=res
                )

            try:
                req.set_result(response)
            except asyncio.InvalidStateError:
                pass

    async def _exec_loop(self):
        while self.running:
            await asyncio.sleep(self.execute_pause)

            requests = []

            for _ in range(25):
                if not self.requests_queue:
                    break

                requests.append(self.requests_queue.pop(0))

            if not requests:
                continue

            code = "return ["

            for r in requests:
                code += "API.{}({}),".format(
                    r.method, json.dumps(r.kwargs, ensure_ascii=False)
                )

            code += "];"

            asyncio.ensure_future(self._exec_perform(code, requests))

    async def update_longpoll_data(self):
        """Update manager's longpoll data"""

        longpoll = await self.raw_request(
            "groups.getLongPollServer", group_id=self.group_id
        )

        if longpoll.error:
            raise ValueError(
                "Couldn't get longpoll information\n{}".format(
                    longpoll.errors
                )
            )

        self.longpoll = {
            **longpoll.response
        }

    async def receiver(self):
        """Return new updates for bot from vkontakte."""

        longpoll_url = self.longpoll_url_template.format(
            self.longpoll["server"],
            self.longpoll["key"],
            self.longpoll["ts"],
        )

        try:
            async with self.session.post(longpoll_url) as resp:
                response = await resp.json()

        except (json.JSONDecodeError, aiohttp.ClientError):
            return ()

        if "ts" in response:
            self.longpoll["ts"] = response["ts"]

        if "failed" in response:
            if response["failed"] in (2, 3, 4):
                await self.update_longpoll_data()

            return ()

        updates = []

        for update in response["updates"]:
            if "type" not in update or "object" not in update:
                continue

            updates.append(update)

        return updates

    async def get_receiver_coroutine_function(self):
        if not self.session:
            self.session = aiohttp.ClientSession()

        current_group_s = await self.raw_request("groups.getById")

        if current_group_s.error or not current_group_s.response[0].get("id"):
            raise ValueError(
                "Couldn't get group information\n{}"
                .format(
                    current_group_s.errors
                )
            )

        self.group_id = current_group_s.response[0]["id"]

        await self.raw_request(
            "groups.setLongPollSettings",
            group_id=self.group_id,
            api_version=self.version,
            enabled=1,
            **{
                **dict(
                    message_new=1, message_reply=0, message_allow=0,
                    message_deny=0, message_edit=0, photo_new=0, audio_new=0,
                    video_new=0, wall_reply_new=0, wall_reply_edit=0,
                    wall_reply_delete=0, wall_reply_restore=0, wall_post_new=0,
                    wall_repost=0, board_post_new=0, board_post_edit=0,
                    board_post_restore=0, board_post_delete=0, photo_comment_new=0,
                    photo_comment_edit=0, photo_comment_delete=0,
                    photo_comment_restore=0, video_comment_new=0,
                    video_comment_edit=0, video_comment_delete=0,
                    video_comment_restore=0, market_comment_new=0,
                    market_comment_edit=0, market_comment_delete=0,
                    market_comment_restore=0, poll_vote_new=0, group_join=0,
                    group_leave=0, group_change_settings=0, group_change_photo=0,
                    group_officers_edit=0, user_block=0, user_unblock=0
                ),
                **self.longpoll_settings
            }
        )

        await self.update_longpoll_data()

        logger.info(
            "logged in as \"%s\" ( https://vk.com/club%s )",
            current_group_s.response[0]["name"],
            current_group_s.response[0]["id"]
        )

        return self.receiver

    async def startup(self, application):
        if self.running:
            return

        self.running = True

        asyncio.ensure_future(self._exec_loop())

    async def dispose(self):
        self.running = False

        if self.session:
            await self.session.close()



"""
Manager, environment and utils for interacting with VKontakte.

Example of initializating VKManager:

.. code-block:: python

    manager1 = VKManager(
        vk_token
    )

"""

from .manager import VKManager, VKRequest, VKResponse
from .environment import VKEnvironment



import asyncio
import json
import logging
import types
import unittest

import aiohttp
from kutana import (Attachment, ExitException, Kutana, Plugin, TGManager,
                    TGResponse)
from kutana.manager.tg.environment import TGAttachmentTemp


class TestManagerVk(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.get_event_loop()
        logging.disable(logging.INFO)

    def tearDown(self):
        logging.disable(logging.DEBUG)

    def test_tg_exceptions(self):
        with self.assertRaises(ValueError):
            TGManager("")

    def test_tg_startup(self):
        mngr = TGManager("token")

        self.assertEqual(
            self.loop.run_until_complete(mngr.startup(None)), None
        )

    def test_tg_environmnet(self):
        mngr = TGManager("token")

        env = self.loop.run_until_complete(
            mngr.get_environment({"message": {"chat": {"id": 1}}})
        )

        self.assertEqual(env.peer_id, 1)

        env = self.loop.run_until_complete(
            mngr.get_environment({})
        )

        self.assertEqual(env.peer_id, None)

    def test_tg_upload_photo(self):
        mngr = TGManager("token")

        env = self.loop.run_until_complete(
            mngr.get_environment({"message": {"chat": {"id": 1}}})
        )

        photo_a = self.loop.run_until_complete(
            env.upload_photo("photo")
        )

        self.assertEqual(photo_a.type, "photo")
        self.assertEqual(photo_a.content, "photo")
        self.assertEqual(photo_a.kwargs, {})

    def test_tg_upload_doc(self):
        mngr = TGManager("token")

        env = self.loop.run_until_complete(
            mngr.get_environment({"message": {"chat": {"id": 1}}})
        )

        document_a = self.loop.run_until_complete(
            env.upload_doc("document")
        )

        self.assertEqual(document_a.type, "doc")
        self.assertEqual(document_a.content, "document")
        self.assertEqual(document_a.kwargs, {})

    def test_tg_replay(self):
        mngr = TGManager("token")

        env = self.loop.run_until_complete(
            mngr.get_environment({"message": {"chat": {"id": 1}}})
        )

        async def request(_, method, **kwargs):
            self.assertEqual(kwargs["chat_id"], "1")

            return TGResponse(True, (), "")

        mngr.request = types.MethodType(request, mngr)

        res = self.loop.run_until_complete(
            env.reply("hi")
        )

        self.assertTrue(res[0].error)
        self.assertFalse(res[0].response)

    def test_tg_get_file_from_attachment(self):
        mngr = TGManager("token")

        env = self.loop.run_until_complete(
            mngr.get_environment({"message": {"chat": {"id": 1}}})
        )

        async def request(_, method, **kwargs):
            self.assertEqual(kwargs["file_id"], 13)

            return TGResponse(False, (), {"file_path": "path"})

        mngr.request = types.MethodType(request, mngr)

        async def request_file(_, path):
            self.assertEqual(path, "path")

            return "file"

        mngr.request_file = types.MethodType(request_file, mngr)

        attachment = Attachment("photo", 13, None, None, None, None)

        file = self.loop.run_until_complete(
            env.get_file_from_attachment(attachment)
        )

        self.assertEqual(file, "file")

    def test_tg_get_file_from_attachment_error(self):
        mngr = TGManager("token")

        env = self.loop.run_until_complete(
            mngr.get_environment({"message": {"chat": {"id": 1}}})
        )

        async def request(_, method, **kwargs):
            self.assertEqual(kwargs["file_id"], 13)

            return TGResponse(True, (), "")

        mngr.request = types.MethodType(request, mngr)

        attachment = Attachment("photo", 13, None, None, None, None)

        file = self.loop.run_until_complete(
            env.get_file_from_attachment(attachment)
        )

        self.assertEqual(file, None)

    def test_tg_request_file_none(self):
        mngr = TGManager("token")

        mngr.file_url = "wrong_url"

        res = self.loop.run_until_complete(
            mngr.request_file("path")
        )

        self.assertEqual(res, None)

        self.loop.run_until_complete(
            mngr.dispose()
        )

    def test_tg_send_message(self):
        mngr = TGManager("token")

        attachment1 = Attachment("photo", 13, None, None, None, None)
        attachment2 = TGAttachmentTemp("bad_type", "strange_content", {})

        async def request(self, method, **kwargs):
            return TGResponse(
                False, (), [method, kwargs]
            )

        mngr.request = types.MethodType(request, mngr)

        res0 = self.loop.run_until_complete(
            mngr.send_message("hi", None)
        )

        self.assertEqual(len(res0), 0)

        res1 = self.loop.run_until_complete(
            mngr.send_message("hi", 1, [attachment1, attachment2])
        )

        self.assertEqual(len(res1), 2)

        self.assertEqual(
            res1[0].response, ["sendMessage", {"chat_id": '1', "text": "hi"}]
        )

        self.assertEqual(
            res1[1].response, ["sendPhoto", {"chat_id": '1', "photo": "13"}]
        )

        res2 = self.loop.run_until_complete(
            mngr.send_message("", 1, attachment1)
        )

        self.assertEqual(len(res2), 1)

        self.assertEqual(
            res2[0].response, ["sendPhoto", {"chat_id": '1', "photo": "13"}]
        )

        self.loop.run_until_complete(
            mngr.dispose()
        )

    def test_tg_create_attachment(self):
        attachment = TGManager.create_attachment(
            {"file_id": 13}, "document"
        )

        self.assertEqual(attachment.type, "doc")
        self.assertEqual(attachment.id, 13)
        self.assertEqual(attachment.link, None)

        attachment = TGManager.create_attachment(
            [{"file_id": 13}], "photo"
        )

        self.assertEqual(attachment.type, "photo")
        self.assertEqual(attachment.id, 13)
        self.assertEqual(attachment.link, None)

        attachment = TGManager.create_attachment(None, "video")

        self.assertEqual(attachment, None)


    def test_tg_create_message(self):
        mngr = TGManager("token")

        message = self.loop.run_until_complete(
            mngr.create_message({"message": {
                "text": "text",
                "from": {"id": 1},
                "chat": {"id": 2},
                "date": 0
            }})
        )

        self.assertEqual(message.text, "text")
        self.assertEqual(message.attachments, ())
        self.assertEqual(message.from_id, 1)
        self.assertEqual(message.peer_id, 2)
        self.assertEqual(message.date, 0)

        message = self.loop.run_until_complete(
            mngr.create_message({})
        )

        self.assertEqual(message, None)

    def test_vk_request(self):
        mngr = TGManager("token")

        async def prepare():
            class FakeSession:
                def post(self, url, proxy, data):
                    class FakePost:
                        def __init__(self, url, proxy, data):
                            pass

                        async def __aenter__(self):
                            class FakeResponse:
                                async def text(self):
                                    return json.dumps({
                                        "ok": True,
                                        "result": "result"
                                    })

                            return FakeResponse()

                        async def __aexit__(self, exc_type, exc, tb):
                            pass

                    return FakePost(url, proxy, data)

                async def close(self):
                    pass

            mngr.session = FakeSession()

        self.loop.run_until_complete(prepare())

        response = self.loop.run_until_complete(
            mngr.request("method", a1="v1", a2="v2")
        )

        self.assertEqual(response.error, False)
        self.assertEqual(response.response, "result")
        self.assertEqual(response.errors, ())

        self.loop.run_until_complete(
            mngr.dispose()
        )

    def test_vk_request_error(self):
        mngr = TGManager("token")

        exception = None

        async def prepare():
            class FakeSession:
                def post(self, url, proxy, data):
                    class FakePost:
                        def __init__(self, url, proxy, data):
                            pass

                        async def __aenter__(self):
                            class FakeResponse:
                                async def text(self):
                                    if exception:
                                        raise exception

                                    return json.dumps({
                                        "ok": False
                                    })

                            return FakeResponse()

                        async def __aexit__(self, exc_type, exc, tb):
                            pass

                    return FakePost(url, proxy, data)

                async def close(self):
                    pass

            mngr.session = FakeSession()

        self.loop.run_until_complete(prepare())

        response = self.loop.run_until_complete(
            mngr.request("method1", a1="v1", a2="v2")
        )

        self.assertEqual(response.error, True)
        self.assertEqual(response.response, "")
        self.assertTrue(response.errors)

        exception = aiohttp.ClientError

        response = self.loop.run_until_complete(
            mngr.request("method2", a1="v1", a2="v2")
        )

        self.assertEqual(response.error, True)
        self.assertEqual(response.response, "")
        self.assertTrue(response.errors)

        self.loop.run_until_complete(
            mngr.dispose()
        )

    def test_tg_receiver(self):
        mngr = TGManager("token")

        async def request(self, method, **kwargs):
            if method == "getUpdates":
                return TGResponse(
                    False, (), [{"update_id": 10}]
                )

        mngr.request = types.MethodType(request, mngr)

        updates = self.loop.run_until_complete(
            mngr.receiver()
        )

        self.assertEqual(updates, [{"update_id": 10}])

        self.loop.run_until_complete(
            mngr.dispose()
        )

    def test_tg_receiver_error(self):
        mngr = TGManager("token")

        async def request(self, method, **kwargs):
            if method == "getUpdates":
                return TGResponse(True, (), "")

        mngr.request = types.MethodType(request, mngr)

        updates = self.loop.run_until_complete(
            mngr.receiver()
        )

        self.assertEqual(updates, ())

        self.loop.run_until_complete(mngr.dispose())

    def test_tg_get_receiver_coroutine_function(self):
        mngr = TGManager("token")

        async def request(self, method, **kwargs):
            if method == "getMe":
                return TGResponse(
                    False,
                    (),
                    {"first_name": "A", "last_name": "B", "username": "a"}
                )

        mngr.request = types.MethodType(request, mngr)

        receiver = self.loop.run_until_complete(
            mngr.get_receiver_coroutine_function()
        )

        self.assertEqual(receiver, mngr.receiver)

        self.loop.run_until_complete(mngr.dispose())

    def test_tg_get_receiver_coroutine_function_error(self):
        mngr = TGManager("token")

        async def request(self, method, **kwargs):
            if method == "getMe":
                return TGResponse(True, (), {})

        mngr.request = types.MethodType(request, mngr)

        with self.assertRaises(ValueError):
            self.loop.run_until_complete(
                mngr.get_receiver_coroutine_function()
            )

        self.loop.run_until_complete(mngr.dispose())



from kutana import Plugin

from testing_tools import KutanaTest


class TestPlugins(KutanaTest):
    def test_plugin_on_text_exception(self):
        plugin = Plugin()

        with self.assertRaises(ValueError):
            plugin.on_text()

    def test_plugin_on_after_processed_exception(self):
        plugin = Plugin()

        with self.assertRaises(ValueError):
            plugin.on_after_processed(messages=False, updates=False)

    def test_plugin_on_after_processed(self):
        plugin = Plugin()

        plugin.on_after_processed(messages=True, updates=False)(1)

        self.assertTrue(plugin._callbacks)
        self.assertFalse(plugin._callbacks_raw)

        plugin = Plugin()

        plugin.on_after_processed(messages=False, updates=True)(1)

        self.assertFalse(plugin._callbacks)
        self.assertTrue(plugin._callbacks_raw)

        plugin = Plugin()

        plugin.on_after_processed(messages=True, updates=True)(1)

        self.assertTrue(plugin._callbacks)
        self.assertTrue(plugin._callbacks_raw)



import asyncio

from kutana import Environment, VKEnvironment, TGEnvironment

from testing_tools import KutanaTest


class TestEnvironments(KutanaTest):
    def test_replace_method(self):
        messages = []

        async def save(self, message):
            messages.append(message)

        env = Environment(None)
        env.replace_method("reply", save)

        asyncio.get_event_loop().run_until_complete(
            env.reply("message")
        )

        self.assertEqual(messages, ["message"])

    def test_spawn(self):
        env = Environment("manager")
        inner_env = env.spawn()

        self.assertEqual(env.manager, inner_env.manager)
        self.assertEqual(env.parent, None)
        self.assertEqual(inner_env.parent, env)

    def test_replace_method_child_environments(self):
        env = Environment(None)

        def method_1(self, message):
            return 1

        def method_2(self, file):
            return 2

        def method_3(self, message):
            return 3

        env.replace_method("reply", method_1)
        env.replace_method("upload_doc", method_2)

        self.assertEqual(env.reply("_"), 1)
        self.assertEqual(env.upload_doc(b"_"), 2)

        child_env = env.spawn()

        child_env.replace_method("reply", method_3)

        self.assertEqual(child_env.reply("_"), 3)
        self.assertEqual(child_env.upload_doc(b"_"), 2)

        child_child_env = child_env.spawn()

        self.assertEqual(child_child_env.reply("_"), 3)
        self.assertEqual(child_child_env.upload_doc(b"_"), 2)

        other_child_env = env.spawn()

        self.assertEqual(other_child_env.reply("_"), 1)
        self.assertEqual(other_child_env.upload_doc(b"_"), 2)

    def test_failed_register_after_processed(self):
        env = Environment("manager")

        with self.assertRaises(ValueError):
            env.register_after_processed()

    def test_failed_method_replace(self):
        env = Environment("manager")

        with self.assertRaises(ValueError):
            env.replace_method("no_method_with_that_name", lambda: 1)

    def test_failed_setattr(self):
        env = Environment("manager")

        with self.assertRaises(AttributeError):
            env.reply = 1

        with self.assertRaises(AttributeError):
            setattr(env, "reply", 1)

    def test_get(self):
        env = Environment("manager")

        self.assertEqual(env.get("attr"), None)
        self.assertEqual(env.get("attr", "default"), "default")

        env.attr = "value"

        self.assertEqual(env.get("attr"), "value")

    def test_concrete_environments(self):
        env1 = VKEnvironment("vk")
        env2 = TGEnvironment("tg")

        env1.replace_method("reply", print)
        env1.a = 10

        child1 = env1.spawn()

        self.assertEqual(child1.a, 10)
        self.assertEqual(child1.reply.func, print)

        env2.replace_method("reply", print)
        env2.a = 10

        child2 = env2.spawn()

        self.assertEqual(child2.a, 10)
        self.assertEqual(child2.reply.func, print)



"""Kutana - engine for developing bots."""

from .environment import Environment
from .exceptions import ExitException
from .loaders import import_module, load_plugins, load_plugins_from_file
from .utils import get_path, is_list_or_tuple, sort_callbacks
from .kutana import Kutana
from .logger import logger, set_logger_level
from .manager.manager import Manager
from .manager.debug import DebugEnvironment, DebugManager
from .manager.tg import TGEnvironment, TGManager, TGResponse
from .manager.vk import VKEnvironment, VKManager, VKRequest, VKResponse
from .plugin import Attachment, Message, Plugin

try:
    import uvloop
    uvloop.install()
except ImportError:  # pragma: no cover
    pass


NAME = "kutana"



from testing_tools import KutanaTest


class TestManagerDebug(KutanaTest):
    def test_plain(self):
        self.target = ["message"]

        with self.debug_manager(["echo message"]) as plugin:

            async def on_echo(message, env):
                await env.reply(env.body)

            plugin.on_startswith_text("echo")(on_echo)

        self.assertEqual(self.manager.replies_all, {1: ["message"]})

    def test_request(self):
        with self.debug_manager(["echo message"]) as plugin:

            async def on_echo(message, env):
                self.assertEqual(await env.request("method"), None)

            plugin.on_startswith_text("echo")(on_echo)


    def test_upload_photo(self):
        with self.debug_manager(["132"]) as plugin:

            async def on_text(message, env):
                attachment = await env.upload_photo("photo")

                self.assertEqual("photo", attachment)
                self.assertEqual(
                    "photo", await env.get_file_from_attachment(attachment)
                )

            plugin.on_has_text()(on_text)

    def test_send_message(self):
        self.target = ["message"]

        with self.debug_manager(["echo message"]) as plugin:
            async def on_echo(message, env):
                await env.send_message(env.body, 1)

            plugin.on_startswith_text("echo")(on_echo)

    def test_multiple_senders_plain(self):
        self.target = ["message1"]

        with self.debug_manager([
            (1, "echo message1"),
            (2, "echo message2"),
            (2, "echo message3"),
            (3, "echo message4")
        ]) as plugin:

            async def on_echo(message, env):
                await env.reply(env.body)

            plugin.on_startswith_text("echo")(on_echo)

        self.assertEqual(
            self.manager.replies_all,
            {1: ["message1"], 2: ["message2", "message3"], 3: ["message4"]}
        )



"""Structures and classes for plugins."""

import re
from collections import namedtuple

from .utils import sort_callbacks


Message = namedtuple(
    "Message",
    "text attachments from_id peer_id date raw_update"
)

Message.__doc__ = """
Text message with possible attachments.

:param text: text message of update
:param attachments: tuple of instances of :class:`.Attachment`
:param from_id: author of sent update (can be None)
:param peer_id: chat identificator or same as from_id (can be None)
:param date: updates's date
:param raw_update: raw update from service
"""


Attachment = namedtuple(
    "Attachment",
    "type id owner_id access_key link raw_attachment"
)

Attachment.__doc__ = """
Detailed information about attachment.

:param type: type of attachment
:param id: attachment's id
:param owner_id: attachment's owner's id
:param access_key: attachment's access key (can be None)
:param link: lint to attachment (can be None)
:param raw_attachment: raw attachment from service
"""


def is_done(value):
    """
    Helper function for deciding if value is sign of successfull
    update's procession.

    :param value: value to interpret
    :returns: True if value is None or equals to "DONE" otherwise False
    """

    return value is None or value == "DONE"


class Plugin:

    """Class for creating extensions for kutana application."""

    def __init__(self, **kwargs):
        self._callbacks = []
        self._callbacks_raw = []

        self._callback_dispose = None
        self._callback_startup = None

        self.priority = 0

        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_callbacks(self):
        """
        Return callbacks for registration in application.

        :returns: list of callbacks to register in application
        """

        return (self._proc_update,)

    def get_callbacks_for_dispose(self):
        """
        Return dispose callbacks for registration in application.

        :returns: list of callbacks to register in application
        """

        return (self._callback_dispose,) if self._callback_dispose else ()

    def get_callbacks_for_startup(self):
        """
        Return startup callbacks for registration in application.

        :returns: list of callbacks to register in application
        """

        return (self._callback_startup,) if self._callback_startup else ()

    async def _proc_update(self, update, env):
        """
        Process update with env and target callbacks.

        :param update: update to process
        :param env: :class:`.Environment` to process update with
        :returns: "DONE" if update is considired updated, None otherwise
        """

        if env.has_message():
            message = env.get_message()

        else:
            message = await env.manager.create_message(update)

            env.set_message(message)

        callbacks = [self._callbacks, self._callbacks_raw]

        if message:
            callbacks_type = 0  # callbacks for messages

        else:
            callbacks_type = 1  # callbacks for raw

        inner_env = env.spawn()

        for _, callback in callbacks[callbacks_type]:
            res = await callback(message if message else update, inner_env)

            if res == "DONE":
                return "DONE"

    def register(self, *callbacks, priority=0):
        """
        Register for processing updates in this plugin.

        :param callbacks: callbacks for registration in this plugin
        :param priority: priority of callbacks **inside** of this plugin
        """

        for callback in callbacks:
            self._callbacks.append((priority, callback))

        sort_callbacks(self._callbacks)

    def register_raw(self, *callbacks, priority=0):
        """
        Register for processing raw updates in this plugin. These callbacks
        will be used if update can't be converted to :class:`.Message`

        :param callbacks: callbacks for registration in this plugin
        :param priority: priority of callbacks **inside** of this plugin
        """

        for callback in callbacks:
            self._callbacks_raw.append((priority, callback))

        sort_callbacks(self._callbacks_raw)

    def on_dispose(self):
        """
        Return decorator for adding callbacks which will be triggered when
        everything is going to shutdown. Only last registered callback will be
        used.

        :returns: decorator for adding callbacks
        """

        def decorator(coro):
            self._callback_dispose = coro

            return coro

        return decorator

    def on_startup(self):
        """
        Return decorator for adding callbacks which will be triggered
        at the startup of kutana. Decorated coroutine receives
        :class:`.Kutana` application as first argument.
        Only last registered callback will be used.

        :returns: decorator for adding callbacks
        """

        def decorator(coro):
            self._callback_startup = coro

            return coro

        return decorator

    def on_raw(self, priority=0):
        """
        Return decorator for adding callback which is triggered
        every time when update can't be turned into :class:`.Message` or
        :class:`.Attachment` object. Arguments raw update and
        :class:`.Environment` is passed to callback.

        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        def decorator(coro):
            async def wrapper(update, env):
                if is_done(await coro(update, env)):
                    return "DONE"

            self.register_raw(wrapper, priority=priority)

            return coro

        return decorator

    def on_text(self, *texts, priority=0):
        """
        Return decorator for adding callback which is triggered
        when the message and any of the specified text are fully matched.

        :param texts: texts to match messages' texts against
        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        if not texts:
            raise ValueError('No texts passed to "Plugin.on_text"')

        check_texts = list(text.strip().lower() for text in texts)

        def decorator(coro):
            async def wrapper(message, env):
                if message.text.strip().lower() in check_texts:
                    if is_done(await coro(message, env)):
                        return "DONE"

            self.register(wrapper, priority=priority)

            return coro

        return decorator

    def on_has_text(self, *texts, priority=0):
        """
        Return decorator for adding callback which is triggered
        when the message contains any of the specified texts.

        Environment's additional fields:

        - `found_text` - text found in message.

        :param texts: texts to search in messages' texts or nothing if you
            want callback to be called on any text
        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        check_texts = tuple(text.strip().lower() for text in texts) or ("",)

        def decorator(coro):
            async def wrapper(msg, env):
                check_text = msg.text.strip().lower()

                if not check_text:
                    return "GOON"

                for text in check_texts:
                    if text not in check_text:
                        continue

                    env.found_text = text

                    if is_done(await coro(msg, env)):
                        return "DONE"

                    break

            self.register(wrapper, priority=priority)

            return coro

        return decorator

    def on_message(self, priority=0):
        """
        Return decorator for adding callback which is triggered
        when the message exists.

        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        def decorator(coro):
            async def wrapper(msg, env):
                if is_done(await coro(msg, env)):
                    return "DONE"

            self.register(wrapper, priority=priority)

            return coro

        return decorator

    def on_startswith_text(self, *texts, priority=0):
        """
        Return decorator for adding callbacks which is triggered
        when the message starts with any of the specified texts.

        Environment's additional fields:

        - `body` - text after prefix.
        - `args` - text after prefix splitted by spaces.
        - `prefix` - text before body.

        :param texts: texts to search in messages' texts begginngs
        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        check_texts = tuple(text.lstrip().lower() for text in texts)

        def decorator(coro):
            def search_prefix(message):
                for text in check_texts:
                    if message.startswith(text):
                        return text

                return None

            async def wrapper(msg, env):
                search_result = search_prefix(msg.text.lower())

                if search_result is None:
                    return

                env.body = msg.text[len(search_result):].strip()
                env.args = env.body.split()
                env.prefix = msg.text[:len(search_result)].strip()

                if is_done(await coro(msg, env)):
                    return "DONE"

            self.register(wrapper, priority=priority)

            return coro

        return decorator

    def on_regexp_text(self, regexp, flags=0, priority=0):
        """
        Returns decorator for adding callback which is triggered
        when the message matches the specified regular expression.

        Environment's additional fields:

        - `match` - :class:`re.Match` object for message.

        :param regexp: regular expression to match messages' texts with
        :param flags: flags for :func:`re.compile`
        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        if isinstance(regexp, str):
            compiled = re.compile(regexp, flags=flags)

        else:
            compiled = regexp

        def decorator(coro):
            async def wrapper(message, env):
                match = compiled.match(message.text)

                if not match:
                    return

                env.match = match

                if is_done(await coro(message, env)):
                    return "DONE"

            self.register(wrapper, priority=priority)

            return coro

        return decorator

    def on_attachment(self, *types, priority=0):
        """
        Returns decorator for adding callback which is triggered
        when the message has attachments of the specified type
        (if no types specified, then any attachments).

        :param types: attachments' types to look for in message
        :param priority: priority of callbacks **inside** of this plugin
        :returns: decorator for adding callback
        """

        def decorator(coro):
            async def wrapper(message, env):
                attachments = message.attachments

                if not attachments:
                    return

                if types:
                    for a in attachments:
                        if a.type in types:
                            break
                    else:
                        return

                if is_done(await coro(message, env)):
                    return "DONE"

            self.register(wrapper, priority=priority)

            return coro

        return decorator

    def on_after_processed(self, messages=True, updates=False):
        """
        Returns decorator for adding callback which is triggered
        when the message or update has been processed. If environment contains
        attribute "exception", that means exception was raised while
        other callbacks was processed. This callback has priority 1000000

        Callbacks, registered with `on_after_processed` should not raise
        any exceptions.

        :param messages: call callback for messages
        :param updates: call callback for updates

        :returns: decorator for adding callback
        """

        if not messages and not updates:
            raise ValueError(
                "`messages` or `updates` must be True for `on_after_processed`"
            )

        def decorator(coro):
            async def wrapper(_, env):
                env.parent.register_after_processed(coro)

                return "GOON"

            if messages:
                self.register(wrapper, priority=1000000)

            if updates:
                self.register_raw(wrapper, priority=1000000)

            return coro

        return decorator



"""
Manager, environment and utils for interacting with Telegram.

Proxy should be in format "http://address_of_prox:port".

Example of initializating TGManager:

.. code-block:: python

    manager2 = TGManager(
        tg_token,
        proxy=tg_proxy,
    )

.. note::
    - Attachmnet type "document" is replaced with "doc" inside of application.
    - Files and photos can't be uploaded without "send_message" or "replay".

"""

from .manager import TGManager, TGResponse
from .environment import TGEnvironment



"""Manager for interacting with telegram."""

import asyncio
import json
from collections import namedtuple

import aiohttp
from kutana.logger import logger
from kutana.manager.manager import Manager
from kutana.manager.tg.environment import TGAttachmentTemp, TGEnvironment
from kutana.plugin import Attachment, Message


TGResponse = namedtuple(
    "TGResponse",
    "error errors response"
)

TGResponse.__doc__ = """
Response from telegram.

:param error: boolean value indicating if error happened
:param errors: array with happened errors
:param response: result of reqeust if no errors happened
"""


class TGManager(Manager):

    """
    Class for receiving updates from telegram.

    :param token: bot's token
    :param messages_per_seconds: how many messages per second bot can send
    :param proxy: proxy to use for bot's requests
    """


    _type = "telegram"


    def __init__(self, token, messages_per_second=29, proxy=None):
        if not token:
            raise ValueError('No "token" specified')

        self.session = None

        self.offset = 0
        self.proxy = proxy

        self.send_messages_pause = 1 / messages_per_second
        self.send_messages_lock = asyncio.Lock()

        self.token = token

        self.api_url = \
            "https://api.telegram.org/bot{}/{{}}".format(self.token)
        self.file_url = \
            "https://api.telegram.org/file/bot{}/{{}}".format(self.token)

    async def get_environment(self, update):
        if "message" in update:
            peer_id = update["message"]["chat"]["id"]

        else:
            peer_id = None

        return TGEnvironment(self, peer_id=peer_id)

    async def request_file(self, path):
        """
        Download file with specified path and return it as bytes.

        :param path: Telegram's file's path for downloading
        :returns: bytes or None
        """

        if not self.session:
            self.session = aiohttp.ClientSession()

        try:
            async with self.session.get(self.file_url.format(path),
                                        proxy=self.proxy) as response:

                return await response.read()

        except aiohttp.ClientError:
            return None

    async def request(self, method, **kwargs):
        """
        Perform request to telegram and return result.

        :param method: method to call
        :param kwargs: parameters for method
        :returns: :class:`.TGResponse`
        """

        if not self.session:
            self.session = aiohttp.ClientSession()

        data = {k: v for k, v in kwargs.items() if v is not None}

        try:
            async with self.session.post(self.api_url.format(method),
                                         proxy=self.proxy,
                                         data=data) as response:

                raw_response_text = await response.text()

                raw_response = json.loads(raw_response_text)

        except (json.JSONDecodeError, aiohttp.ClientError) as e:
            return TGResponse(
                error=True,
                errors=(("Kutana", str(type(e)) + ": " + str(e)),),
                response=""
            )

        if not raw_response["ok"]:
            return TGResponse(
                error=True,
                errors=(
                    ("TG_err", (
                        raw_response.get("error_code", ""),
                        raw_response.get("description", "")
                    ))
                ),
                response=""
            )

        return TGResponse(
            error=False,
            errors=(),
            response=raw_response["result"]
        )

    async def send_message(self, message, peer_id, attachment=None, **kwargs):
        """
        Send message to target peer_id with parameters.

        :param message: text to send
        :param peer_id: target recipient
        :param attachment: list of instances of :class:`.Attachment` or
            :class:`.TGAttachmentTemp`
        :parma kwargs: arguments to send to telegram's `sendMessage`
        :returns: list of responses from telegram
        """

        await self.send_messages_lock.acquire()

        try:
            return await self._send_message(
                message, peer_id, attachment, **kwargs
            )

        finally:
            await asyncio.sleep(self.send_messages_pause)

            self.send_messages_lock.release()

    async def _send_message(self, message, peer_id, attachment=None, **kwargs):
        """
        Send message to target peer_id with parameters without checking for
        limits.

        :param message: text to send
        :param peer_id: target recipient
        :param attachment: list of instances of :class:`.Attachment` or
            :class:`.TGAttachmentTemp`
        :parma kwargs: arguments to send to telegram's `sendMessage`
        :returns: list of responses from telegram
        """

        if peer_id is None:
            return ()

        peer_id_str = str(peer_id)

        result = []

        if message:
            message_parts = self.split_large_text(message)

            for part in message_parts:
                result.append(
                    await self.request(
                        "sendMessage", chat_id=peer_id_str, text=part, **kwargs
                    )
                )

        if isinstance(attachment, (Attachment, TGAttachmentTemp)):
            attachment = [attachment]

        if isinstance(attachment, (list, tuple)):
            new_attachment = []

            for a in attachment:
                if isinstance(a, Attachment):
                    attachment_type = "document" if a.type == "doc" else a.type

                    new_attachment.append((attachment_type, str(a.id), {}))

                elif isinstance(a, TGAttachmentTemp):
                    new_attachment.append((a.type, a.content, a.kwargs))

            attachment = new_attachment

        if attachment:
            for attachment_type, content, upload_kwargs in attachment:
                if attachment_type not in ("document", "photo", "audio",
                                           "video", "voice"):
                    continue

                result.append(
                    await self.request(
                        "send" + attachment_type.capitalize(),
                        **{
                            "chat_id": peer_id_str,
                            attachment_type: content,
                            **upload_kwargs
                        }
                    )
                )

        return result

    async def create_message(self, update):
        if "message" not in update:
            return None

        attachments = []

        for key in ("audio", "photo", "video", "document"):
            attachments.append(
                self.create_attachment(update["message"].get(key), key)
            )

        return Message(
            update["message"].get("text", ""),
            tuple(a for a in attachments if a),
            update["message"]["from"]["id"],
            update["message"]["chat"]["id"],  # peer_id
            update["message"]["date"],
            update
        )

    @staticmethod
    def create_attachment(attachment, attachment_type=None):
        """
        Create and return :class:`.Attachment` created from passed data. If
        attachment type can't be determined passed "attachment_type"
        well be used.
        """

        if not attachment:
            return None

        if attachment_type == "document":
            attachment_type = "doc"

        elif attachment_type == "photo":
            attachment = attachment[-1]

        return Attachment(
            attachment_type,
            attachment.get("file_id"),
            None,
            None,
            None,
            attachment
        )

    async def receiver(self):
        """Return new updates for bot from vkontakte."""

        response = await self.request(
            "getUpdates", timeout=25, offset=self.offset
        )

        if response.error or not response.response:
            return ()

        updates = response.response

        self.offset = updates[-1]["update_id"] + 1

        return updates

    async def get_receiver_coroutine_function(self):
        if not self.session:
            self.session = aiohttp.ClientSession()

        response = await self.request("getMe")

        if response.error:
            raise ValueError("Wrong token for TGManager")

        logger.info(
            "logged in as \"%s\" ( https://t.me/%s )",
            (
                response.response.get("first_name", "") + " " +
                response.response.get("last_name", "")
            ).strip(),
            response.response["username"]
        )

        return self.receiver

    async def startup(self, application):
        pass

    async def dispose(self):
        if self.session:
            await self.session.close()



import asyncio
import re

from kutana import Attachment, DebugEnvironment, Message, Plugin

from testing_tools import KutanaTest


class TestPlugins(KutanaTest):
    def test_plugin_creation(self):
        plugin = Plugin(name="Name", cmds=["cmd1", "cmd2"])

        self.assertEqual(plugin.name, "Name")  # pylint: disable=E1101
        self.assertEqual(plugin.cmds, ["cmd1", "cmd2"])  # pylint: disable=E1101

        plugin = Plugin(priority=1000)

    def test_echo_plugin(self):
        queue = ["message", "echo message", "echonotecho"]

        self.target = ["message"]

        with self.debug_manager(queue) as plugin:

            async def on_echo(message, env):
                await env.reply(env.body)

            plugin.on_startswith_text("echo ", "echo\n")(on_echo)

    def test_plugin_on_startup(self):
        self.called = 0

        with self.debug_manager(["message"]) as plugin:

            async def startup(app):
                self.called += 1

            plugin.on_startup()(startup)

        self.assertEqual(self.called, 1)

    def test_plugin_register(self):
        self.answers = []

        with self.debug_manager(["message"]) as plugin:

            async def cb_1(update, env):
                self.answers.append(1)

            async def cb_2(update, env):
                self.answers.append(2)

            async def cb_3(update, env):
                self.answers.append(3)

            plugin.register(cb_1)
            plugin.register(cb_2, priority=1)
            plugin.register(cb_3)

        self.assertEqual(self.answers, [2, 1, 3])

    def test_plugin_register_two_plugins(self):
        self.answers = []

        with self.debug_manager(["message"]) as plugin1:

            plugin2 = Plugin(priority=1)

            self.plugins.append(plugin2)

            async def cb_1(update, env):
                self.answers.append(1)

            async def cb_2(update, env):
                self.answers.append(2)

            async def cb_3(update, env):
                self.answers.append(3)

            plugin1.register(cb_1)
            plugin2.register(cb_3)
            plugin1.register(cb_2, priority=1)

        self.assertEqual(self.answers, [3, 2, 1])

    def test_args_on_startswith_text(self):
        queue = ["pr a b c", "pr a \"b c\"", "pr ab c"]

        queue_answer = [
            ["a", "b", "c"], ["a", "\"b", "c\""], ["ab", "c"]
        ]

        with self.debug_manager(queue) as plugin:

            async def on_startswith_text(message, env):
                self.assertEqual(env.args, queue_answer.pop(0))

            plugin.on_startswith_text("pr")(on_startswith_text)

        self.assertFalse(queue_answer)

    def test_environments(self):
        self.target = ["1", "2"]

        with self.debug_manager(["123", ".hi arg1 arg2"]) as plugin:

            async def on_has_text(msg, env):
                self.assertEqual(env.found_text, "12")

                await env.reply("1")

            plugin.on_has_text("12")(on_has_text)


            async def on_startswith_text(msg, env):
                self.assertEqual(env.args, ["hi", "arg1", "arg2"])
                self.assertEqual(env.body, "hi arg1 arg2")
                self.assertEqual(env.prefix, ".")

                await env.reply("2")

            plugin.on_startswith_text(".")(on_startswith_text)

    def test_plugins_callbacks_done(self):
        self.counter = 0

        with self.debug_manager(["123"]) as plugin:

            async def on_has_text(message, env):
                self.counter += 1

                return "DONE"

            plugin.on_has_text("123")(on_has_text)

            async def on_has_text_2(message, env):
                self.counter += 1

            plugin.on_has_text("123")(on_has_text_2)

        self.assertEqual(self.counter, 1)

    def test_plugins_callbacks_not_done(self):
        self.counter = 0

        with self.debug_manager(["123"]) as plugin:

            async def on_has_text(message, env):
                self.counter += 1

                return "GOON"

            plugin.on_has_text("123")(on_has_text)

            async def on_has_text_2(message, env):
                self.counter += 1

            plugin.on_has_text("123")(on_has_text_2)

        self.assertEqual(self.counter, 2)

    def test_multiple_plugins(self):
        self.counter = 0

        with self.debug_manager(["message"]):

            self.plugins.append(Plugin())
            self.plugins.append(Plugin())

            async def on_has_text(message, env):
                self.counter += 1
                return "GOON"

            for pl in self.plugins:
                pl.on_has_text()(on_has_text)

        self.assertEqual(self.counter, 3)

    def test_early_callbacks(self):
        self.answers = []

        with self.debug_manager(["message"]):

            async def on_has_text_early(message, env):
                self.answers.append("early")

                return "GOON"

            async def on_has_text(message, env):
                self.answers.append("late")

            self.plugin.on_has_text()(on_has_text)
            self.plugin.on_has_text(priority=1)(on_has_text_early)

        self.assertEqual(self.answers, ["early", "late"])

    def test_early_callbacks_raw(self):
        self.answers = []

        with self.debug_manager([None]):

            async def on_raw_early(update, env):
                self.answers.append("early")

                return "GOON"

            async def on_raw(update, env):
                self.answers.append("late")

            self.plugin.on_raw()(on_raw)
            self.plugin.on_raw(priority=1)(on_raw_early)

        self.assertEqual(self.answers, ["early", "late"])

    def test_plugin_callbacks(self):
        self.disposed = 0
        self.counter = 0

        with self.debug_manager(["123", "321"]) as plugin:
            async def on_123(message, env):
                self.assertEqual(message.text, "123")
                self.counter += 1

            plugin.on_text("123")(on_123)


            async def on_everything(message, env):
                self.assertNotEqual(message.text, "123")
                self.counter += 1

            plugin.on_has_text()(on_everything)


            async def on_raw(update, env):
                raise RuntimeError

            plugin.on_raw()(on_raw)


            async def on_dispose():
                self.disposed += 1

            plugin.on_dispose()(on_dispose)

        self.assertEqual(self.counter, 2)
        self.assertEqual(self.disposed, 1)

    def test_reply(self):
        self.target = ["echo message"]

        with self.debug_manager(self.target) as plugin:

            async def on_echo(message, env):
                await env.reply(message.text)

            plugin.on_startswith_text("echo")(on_echo)

    def test_plugin_no_attachments_type(self):
        plugin = Plugin()

        async def on_attachment(message, env):
            return "DONE"

        plugin.on_attachment("photo")(on_attachment)

        wrapper = plugin._callbacks[0][1]

        attachments = [
            Attachment("audio", 0, 0, 0, 0, {}),
            Attachment("video", 0, 0, 0, 0, {})
        ]

        res = asyncio.get_event_loop().run_until_complete(
            wrapper(
                Message("", attachments, 0, 0, 0, {}),
                DebugEnvironment(None, 0)
            )
        )

        self.assertEqual(res, None)

    def test_plugin_attachments(self):
        plugin = Plugin()

        async def on_attachment(message, env):
            return "DONE"

        plugin.on_attachment()(on_attachment)

        wrapper = plugin._callbacks[0][1]

        res = asyncio.get_event_loop().run_until_complete(
            wrapper(
                Message("text", ("attachment"), 0, 0, 0, {}),
                DebugEnvironment(None, 0)
            )
        )

        self.assertEqual(res, "DONE")

    def test_plugin_attachments_type(self):
        plugin = Plugin()

        async def on_attachment(message, env):
            return "DONE"

        plugin.on_attachment("photo")(on_attachment)

        wrapper = plugin._callbacks[0][1]

        res = asyncio.get_event_loop().run_until_complete(
            wrapper(
                Message(
                    "text", [Attachment("photo", 0, 0, 0, 0, {})], 0, 0, 0, {}
                ),
                DebugEnvironment(None, 0)
            )
        )

        self.assertEqual(res, "DONE")

    def test_plugin_no_text(self):
        plugin = Plugin()

        async def on_has_text(message, env):
            return "DONE"

        plugin.on_has_text()(on_has_text)

        wrapper = plugin._callbacks[0][1]

        res = asyncio.get_event_loop().run_until_complete(
            wrapper(
                Message("", ("attachment"), 0, 0, 0, {}),
                DebugEnvironment(None, 0)
            )
        )

        self.assertNotEqual(res, "DONE")

    def test_on_message(self):
        plugin = Plugin()

        async def on_message(message, env):
            return "DONE"

        plugin.on_message()(on_message)

        wrapper = plugin._callbacks[0][1]

        res = asyncio.get_event_loop().run_until_complete(
            wrapper(
                Message("", ("attachment"), 0, 0, 0, {}),
                DebugEnvironment(None, 0)
            )
        )

        self.assertEqual(res, "DONE")

    def test_plugin_onstar(self):
        queue = ["привет", "отлично привет", "ecHo", "ab", "ae", "hello"]
        self.result = 0

        with self.debug_manager(queue) as plugin:

            async def no_trigger(message, env):
                self.assertTrue(False)

            plugin.on_attachment("photo", "audio")(no_trigger)


            async def zero_trigger(message, env):
                self.result |= 1 << 0

            plugin.on_has_text("привет")(zero_trigger)


            async def one_trigger(message, env):
                self.assertTrue(env.match.group(0), "ab")

                self.result |= 1 << 1

            plugin.on_regexp_text(r"a(b|c)")(one_trigger)


            async def two_trigger(message, env):
                self.result |= 1 << 2

            plugin.on_regexp_text(re.compile(r"a(e|z)"))(two_trigger)


            async def three_trigger(message, env):
                self.result |= 1 << 3

            plugin.on_startswith_text("hell")(three_trigger)


            async def four_trigger(message, env):
                self.result |= 1 << 4

            plugin.on_text("ecHo")(four_trigger)

        self.assertEqual(self.result, (1 << 5) - 1)



"""Module with logger that is used in this application."""


import logging
import logging.handlers


FORMATTER = logging.Formatter('%(asctime)s [ %(levelname)s ] %(message)s')
LEVEL = logging.INFO

handler_file = logging.FileHandler('kutana.log')  # pylint: disable=C0103
handler_file.setLevel(LEVEL)
handler_file.setFormatter(FORMATTER)

handler_stream = logging.StreamHandler()  # pylint: disable=C0103
handler_stream.setLevel(LEVEL)
handler_stream.setFormatter(FORMATTER)

logger = logging.getLogger("kutana")  # pylint: disable=C0103
logger.setLevel(LEVEL)
logger.addHandler(handler_stream)
logger.addHandler(handler_file)


def set_logger_level(level):
    """
    Set logging level for stream, file and logger.

    :param level: logging level to set
    """

    handler_file.setLevel(level)
    handler_stream.setLevel(level)
    logger.setLevel(level)



"""Environment for :class:`.TGManager`."""

from collections import namedtuple

from kutana.environment import Environment


TGAttachmentTemp = namedtuple(
    "TGAttachmentTemp",
    "type content kwargs"
)


TGAttachmentTemp.__doc__ = """
Tricky namedtuple for storing not yet uploaded files.

:param type: attachment's type
:param content: attachment's content
:param kwargs: kwargs for telegram's methods
"""


class TGEnvironment(Environment):

    """Environment for :class:`.TGManager`"""

    async def reply(self, message, attachment=None, **kwargs):
        """
        Reply to currently processed message. If text is too long - message
        will be splitted into parts.

        :param message: message to reply with
        :param attachment: optional attachment or list of attachments to
            reply with
        :param kwargs: arguments for telegrams's "sendMessage",
            "sendPhoto" etc.
        :returns: list with results of sending messages
        """

        return await self.manager.send_message(
            message, self.peer_id, attachment, **kwargs
        )

    async def upload_doc(self, file, **kwargs):
        """
        Pack file to be sent with :func:`.send_message`
        (or :func:`.reply`) as document.

        :param file: document as file or bytes
        :param kwargs: arguments for telegram's "sendFile"
        :returns: :class:`.TGAttachmentTemp`
        """

        return TGAttachmentTemp("doc", file, kwargs)

    async def upload_photo(self, file, **kwargs):
        """
        Pack file to be sent with :func:`.send_message`
        (or :func:`.reply`) as photo.

        :param file: photo as file or bytes
        :param kwargs: arguments for telegram's "sendPhoto"
        :returns: :class:`.TGAttachmentTemp`
        """

        return TGAttachmentTemp("photo", file, kwargs)

    async def get_file_from_attachment(self, attachment):
        file_data = await self.manager.request(
            "getFile", file_id=attachment.id
        )

        if file_data.error:
            return None

        return await self.manager.request_file(
            file_data.response["file_path"]
        )



import asyncio
import logging
import unittest

from kutana import (Kutana, Manager, DebugEnvironment, DebugManager,
                    Plugin, VKEnvironment, VKManager, VKResponse, load_plugins,
                    set_logger_level)


class TestMiscellaneous(unittest.TestCase):
    def test_load_plugins(self):
        loaded_plugins = load_plugins("test/test_plugins/")
        loaded_plugins.sort(key=lambda plugin: plugin.name)

        self.assertEqual(len(loaded_plugins), 3)

        self.assertEqual(loaded_plugins[0].name, "Memory")
        self.assertEqual(loaded_plugins[1].name, "My file")
        self.assertEqual(loaded_plugins[2].name, "My file twice")

        app = Kutana()
        app.register_plugins(loaded_plugins)

        loop = asyncio.get_event_loop()

        loop.run_until_complete(app.startup())

        loop.run_until_complete(
            app.process(
                DebugManager(), "message"
            )
        )

        self.assertEqual(loaded_plugins[0].memory, "message")
        self.assertEqual(loaded_plugins[1].my_file, ":)")
        self.assertEqual(loaded_plugins[2].my_file, ":):)")

    def test_split_large_text(self):
        split = Manager.split_large_text

        self.assertEqual(split("abc"), ("abc",))

        parts = split("abc" * 4096)

        self.assertEqual(parts[0], ("abc" * 4096)[:4096])
        self.assertEqual(len(parts), 3)

    def test_plugin_with_exception_callback(self):
        plugin = Plugin(exceptions=0)

        async def on_processed(message, env):
            if env.exception:
                plugin.exceptions += 1

        plugin.on_after_processed()(on_processed)

        async def on_has_text(message, env):
            raise Exception

        plugin.on_has_text()(on_has_text)

        app = Kutana()

        app.register_plugins([plugin])

        loop = asyncio.get_event_loop()

        set_logger_level(logging.CRITICAL)

        loop.run_until_complete(
            app.process(
                DebugManager(), "message"
            )
        )

        set_logger_level(logging.ERROR)

        self.assertEqual(plugin.exceptions, 1)

    def test_vk_conversation(self):
        class FakeManager(VKManager):
            def __init__(self):
                pass

            async def request(self, method, **kwargs):
                return VKResponse(False, (), {"object_id": 1})

        env = VKEnvironment(FakeManager(), peer_id=0)

        loop = asyncio.get_event_loop()

        message = loop.run_until_complete(
            env.manager.create_message(
                {"object": {"date": 1, "random_id": 0, "fwd_messages": [],
                "important": False, "peer_id": 1,
                "text": "echo [club1|\u0421\u043e] 123", "attachments": [],
                "conversation_message_id": 1411, "out": 0, "from_id": 1,
                "id": 0, "is_hidden": False}, "group_id": 1,
                "type": "message_new"}
            )
        )

        self.assertEqual(message.text, "echo  123")
        self.assertEqual(message.attachments, ())

        message = loop.run_until_complete(
            env.manager.create_message(
                {"object": {"date": 1, "random_id": 0, "fwd_messages": [],
                "important": False, "peer_id": 1,
                "text": "echo [club1|\u0421\u043e] 123", "attachments": [],
                "conversation_message_id": 1411, "out": 0, "from_id": 1,
                "id": 0, "is_hidden": False}, "group_id": 2,
                "type": "message_new"}
            )
        )

        self.assertEqual(message.text, "echo [club1|\u0421\u043e] 123")
        self.assertEqual(message.attachments, ())


if __name__ == '__main__':
    unittest.main()



"""Utility functions for different purposes."""

from os.path import dirname, join


def sort_callbacks(callbacks):
    """
    Sort passed list of callbacks by it's first argument (priority) in
    descending order.

    :param callbacks: list of callbacks
    """

    callbacks.sort(key=lambda p: -p[0])


def is_list_or_tuple(obj):
    """
    Return True if object is list or tuple.

    :param obj: object to check
    :returns: True or False
    """

    return isinstance(obj, (list, tuple))


def get_path(root, path):
    """
    Shortcut for ``join(dirname(root), path)``.

    :param root: root path
    :param path: path to file or folder
    :returns: path to file or folder relative to root
    """

    return join(dirname(root), path)



"""Core class for engine."""

import asyncio

from .exceptions import ExitException
from .logger import logger
from .utils import sort_callbacks


class Kutana:

    """
    Main class for constructing application.

    :param loop: loop for working
    """

    def __init__(self, loop=None):
        self.running = True  #: True if application is running
        self.config = {}  #: user's configuration

        self.registered_plugins = []  #: plugins registered in application

        self._callbacks = []
        self._startup_callbacks = []
        self._dispose_callbacks = []

        self._managers = []
        self._loop = loop or asyncio.get_event_loop()
        self._tasks = []
        self._tasks_loops = []

    def register_startup(self, *callbacks, priority=0):
        """
        Register callbacks for startup.

        :param callbacks: callbacks for registration
        :param priority: priority of callbacks
        """

        for callback in callbacks:
            self._startup_callbacks.append((priority, callback))

        sort_callbacks(self._startup_callbacks)

    def register_dispose(self, *callbacks, priority=0):
        """
        Register callbacks for disposing resourses before shutting down.

        :param callbacks: callbacks for registration
        :param priority: priority of callbacks
        """

        for callback in callbacks:
            self._dispose_callbacks.append((priority, callback))

        sort_callbacks(self._dispose_callbacks)

    def register(self, *callbacks, priority=0):
        """
        Register callbacks for processing updates with specified priority.

        :param callbacks: callbacks for registration
        :param priority: priority of callbacks
        """

        for callback in callbacks:
            self._callbacks.append((priority, callback))

        sort_callbacks(self._callbacks)

    def register_plugins(self, plugins):
        """
        Register plugins in application.

        :param plguins: plugins for registration
        """

        for plugin in plugins:
            self.register(
                *plugin.get_callbacks(), priority=plugin.priority
            )

            self.register_startup(
                *plugin.get_callbacks_for_startup(), priority=plugin.priority
            )

            self.register_dispose(
                *plugin.get_callbacks_for_dispose(), priority=plugin.priority
            )

            self.registered_plugins.append(plugin)

    def add_manager(self, manager):
        """Add manager to application."""

        self._managers.append(manager)

    async def process(self, mngr, update):
        """Create environment and process update from manager."""

        env = await mngr.get_environment(update)

        try:
            await env.process(update, self._callbacks)

        except Exception as e:  # pylint: disable=W0703
            await env.reply(
                "Произошла крайне критическая ошибка. Сообщите об этом "
                "администратору."
            )

            logger.exception(e)

    async def loop_for_manager(self, mngr):
        """Receive and process updates from target manager forever."""

        receiver = await mngr.get_receiver_coroutine_function()

        while self.running:
            for update in await receiver():
                task = asyncio.ensure_future(
                    self.process(mngr, update), loop=self._loop
                )

                self._tasks.append(task)

            await asyncio.sleep(0)

    def run(self):
        """Start application."""

        asyncio.set_event_loop(self._loop)

        for manager in self._managers:
            self._tasks_loops.append(self.loop_for_manager(manager))

            self._loop.run_until_complete(
                manager.startup(self)
            )

            # awaitables = self._loop.run_until_complete(
            #     manager.startup(self.ensure_future)
            # )

            # for awaitable in awaitables:
            #     self._tasks_loops.append(awaitable)

        self._loop.run_until_complete(self.startup())

        try:
            self._loop.run_until_complete(asyncio.gather(*self._tasks_loops))

        except (KeyboardInterrupt, ExitException):
            pass

        finally:
            self.running = False

        self._loop.run_until_complete(self.dispose())

    async def startup(self):
        """Prepare plugins for work."""

        for _, callback in self._startup_callbacks:
            await callback(self)


    async def dispose(self):
        """Free resources and prepare for shutdown."""

        await asyncio.gather(*self._tasks)

        for _, callback in self._dispose_callbacks:
            await callback()

        for manager in self._managers:
            await manager.dispose()



"""Base environment for manager."""

import functools
import sys

from .logger import logger


class Environment:

    """Base environment for manager."""

    __slots__ = (
        "manager", "parent", "peer_id",
        "_has_message", "_message",
        "_replaced_methods", "_storage",
        "_istorage"
    )

    def __init__(self, manager, parent_environment=None, peer_id=None):
        self.manager = manager
        self.parent = parent_environment

        self.peer_id = peer_id

        self._replaced_methods = {}
        self._storage = {}
        self._istorage = {
            "cbs_a_p": [],  # callbacks for event "after processed"
        }

        self._has_message = None
        self._message = None

    def __setattr__(self, name, value):
        if name in object.__getattribute__(self, "__slots__"):
            object.__setattr__(self, name, value)

        elif name in dir(self):
            raise AttributeError

        else:
            self._storage[name] = value

    def __getattribute__(self, name):
        replaced_methods = object.__getattribute__(self, "_replaced_methods")

        if name in replaced_methods:
            return functools.partial(replaced_methods[name], self)

        storage = object.__getattribute__(self, "_storage")

        if name in storage:
            return storage[name]

        return object.__getattribute__(self, name)

    def get(self, name, default=None):
        """
        Get attribute of environment of default value if attribute is not
        found.

        :param name: name of the attribute
        :param default: default value
        :returns: value if attribute or default value
        """

        try:
            return self.__getattribute__(name)

        except AttributeError:
            return default

    def replace_method(self, method_name, method):
        """
        Replaces this instance's method with specified name with passed method

        Example:

        .. code-block:: python

            async def replacement(self, message):
                print(message)

            env = Environment(manager)
            env.replace_method("reply", replacement)
            await env.reply("hey")  # will print "hey"

        .. note::

            Many methods are coroutine, so you should replace coroutines with
            coroutines and normal functions with functions.

        :param method_name: method's name to replace
        :param method: method to replace with
        """

        if method_name not in dir(self) \
                or not callable(getattr(self, method_name)):

            raise ValueError("No method with name '{}'".format(method_name))

        self._replaced_methods[method_name] = method

    @property
    def manager_type(self):
        """Return currently processed message's manager type."""

        return self.manager.get_type()

    def spawn(self):
        """
        Create partial copy of environment with this environment as parent.
        """

        new_env = self.__class__(
            manager=self.manager,
            parent_environment=self,
            peer_id=self.peer_id
        )

        for k, v in self._storage.items():
            setattr(new_env, k, v)

        for k, v in self._replaced_methods.items():
            new_env.replace_method(k, v)

        return new_env

    def has_message(self):
        """
        Return True if message was created from update
        (successfully or not).
        """

        return self._has_message

    def get_message(self):
        """Return currently processed message."""

        return self._message

    def set_message(self, message):
        """
        Change message that will be passed to every sequential callbacks'
        calls. Message should be instance of :class:`.Message`
        """

        self._message = message
        self._has_message = message is not None

    async def request(self, method, **kwargs):
        """Proxy for manager's "request" method."""

        return await self.manager.request(
            method, **kwargs
        )

    async def send_message(self, message, peer_id, attachment=None, **kwargs):
        """Proxy for manager's "send_message" method."""

        return await self.manager.send_message(
            message, peer_id, attachment, **kwargs
        )

    async def reply(self, message, attachment=None, **kwargs):
        """
        Reply to currently processed message.

        :param message: message to reply with
        :param attachment: optional attachment or list of attachments
        :param kwargs: arguments for request to service
        :returns: list with results of sending messages
        """

        return await self.manager.send_message(
            message, self.peer_id, attachment, **kwargs
        )

    async def upload_doc(self, file, **kwargs):
        """
        Upload of prepare file to be sent with :func:`.send_message`
        (or :func:`.reply`) as document. This can vary in managers'
        implementations.

        :param file: document as file or bytes
        :param kwargs: arguments for manager's method
        :returns: :class:`.Attachment` or None
        """

        raise NotImplementedError

    async def upload_photo(self, file, **kwargs):
        """
        Upload of prepare file to be sent with :func:`.send_message`
        (or :func:`.reply`) as photo. This can vary in managers'
        implementations.

        :param file: photo as file or bytes
        :param kwargs: arguments for manager's method
        :returns: :class:`.Attachment` or None
        """

        raise NotImplementedError

    async def get_file_from_attachment(self, attachment):
        """
        Try to download passed attachment and return it as bytes. Return None
        if error occured.

        :param attachment: :class:`.Attachment`
        :returns: bytes or None
        """

        raise NotImplementedError

    def register_after_processed(self, *callbacks):
        """
        Add coroutine for calling after update was processed. Coroutines are
        called in order they was added.
        """

        if not callbacks:
            raise ValueError("No callbacks passed")

        self._istorage["cbs_a_p"].extend(callbacks)

    async def process(self, update, callbacks):
        """
        Process update in this environment with passed callbacks.

        :param update: update to process
        :param callbacks: callbacks to process with
        """

        try:
            for _, callback in callbacks:
                if await callback(update, self) == "DONE":
                    break

        except Exception as e:  # pylint: disable=W0703
            logger.exception(
                "\"%s::%s\" on update %s from %s",
                sys.exc_info()[0].__name__, e, update, self.manager_type
            )

            self._storage["exception"] = e

        # Calling callbacks for event "after processed"
        for callback in self._istorage["cbs_a_p"]:
            await callback(update, self)
